// functemp.cpp
// Glenn G. Chappell
// 31 Oct 2013
// Demo of Function Templates

#include <iostream>
using std::cout;
using std::endl;
using std::cin;
using std::ostream;
#include <string>
using std::string;


// superSum (template)
// Computes a+b+a+b for any type that can be added.
template <typename T>
T superSum(T a, T b)
{
    return a + b + a + b;
}


// superSum (double)
// We do something different for double.
// This supersedes the above template when parameters have type double.
double superSum(double a, double b)
{
    return 100. * (a + b);
}


// class Foo
// Silly class with binary op+, stream-insertion op<<
class Foo {
public:

    // Compiler-generated default ctor, copy ctor, copy op=, dctor used

    Foo operator+(const Foo & b) const
    {
        return *this;
    }
};


// operator<< (Foo)
ostream & operator<<(ostream & os, const Foo & f)
{
    os << "Foo!!";
    return os;
}


int main()
{
    cout << "superSum is function template; "
         << "superSum(a, b) returns a + b + a + b" << endl;
    cout << "Except when parameters are double; "
         << "then it returns 100. * (a + b)" << endl;
    cout << endl;
    cout << "Call superSum with various parameter types, "
         << "print the return value:" << endl;
    cout << endl;

    int c = 3;
    int d = 4;
    cout << "int -- parameters are "
         << c << ", " << d << endl;
    cout << superSum(c, d) << endl;
    cout << endl;

    double x = 3.2;
    double y = 4.1;
    cout << "double -- parameters are "
         << x << ", " << y << endl;
    cout << "- function is written, so template not used" << endl;
    cout << superSum(x, y) << endl;
    cout << endl;

    string s1 = "Hi";
    string s2 = "ho";
    cout << "string -- parameters are \""
         << s1 << "\", \"" << s2 << "\"" << endl;
    cout << superSum(s1, s2) << endl;
    cout << endl;

    Foo f1;
    Foo f2;
    cout << "Foo -- parameters are "
         << "Foo(), Foo() [default-constructed Foo objects]" << endl;
    cout << "- Foo op+ does nothing; op<< prints \"Foo!!\"" << endl;
    cout << superSum(f1, f2) << endl;
    cout << endl;

    cout << "Press ENTER to quit ";
    while (cin.get() != '\n') ;

    return 0;
}

